/*************************************************************************
 *
 *  C-to-C Translator
 *
 *  Adapted from Clean ANSI C Parser
 *  Eric A. Brewer, Michael D. Noakes
 * 
 *************************************************************************/
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ast.h"

typedef enum verifycontext {
     TopLevel,
     FormalParm,
     StructField,
     Other
} Context;

PRIVATE void VerifyNode(Node *node, Context c);
PRIVATE void VerifyType(Node *type);
PRIVATE void VerifyDeclList(List *list, Context c);
PRIVATE void VerifyExpr(Node *node);
PRIVATE void VerifyExprList(List *list);
PRIVATE void VerifyTypeOrExpr(Node *node);
PRIVATE void VerifyStmt(Node *node);
PRIVATE void VerifyStmtList(List *list);
PRIVATE void VerifyTq(TypeQual tq);
PRIVATE void VerifyScAndDq(TypeQual tq);

GLOBAL void VerifyParse(List *program)
{
     /* a program is a list of declarations and proc definitions */
     VerifyDeclList(program, TopLevel);
}

/*************************************************************************/
/*                                                                       */
/*                          Expression nodes                             */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void VerifyConst(Node *node, ConstNode *u, Context c)
{
     VerifyType(u->type);
}

PRIVATE inline void VerifyId(Node *node, idNode *u, Context c)
{
     Node *var;

     assert(u->text != NULL);
     var = u->decl;
     /* var can be NULL for struct/union field names, but VerifyExpr
      * is not called for the right side of -> and . */
     assert(var != NULL);
     assert(var->typ == Decl);
     /* name in decl must match id name */
     assert(var->u.decl.name == u->text);
}

PRIVATE inline void VerifyBinop(Node *node, binopNode *u, Context c)
{
     /* fix: check op */
     if (u->op == ARROW || u->op == '.') {
	  VerifyExpr(u->left);
	  assert(u->right != NULL);
	  assert(u->right->typ == Id);
	  assert(u->right->u.id.decl == NULL);
     } else {
	  VerifyExpr(u->left);
	  VerifyExpr(u->right);
     }
}

PRIVATE inline void VerifyUnary(Node *node, unaryNode *u, Context c)
{
     /* fix: check op */
     if (u->op == SIZEOF || u->op == ALIGNOF)
	  VerifyTypeOrExpr(u->expr);
     else
	  VerifyExpr(u->expr);
}

PRIVATE inline void VerifyCast(Node *node, castNode *u, Context c)
{
     VerifyType(u->type);
     VerifyExpr(u->expr);
}

PRIVATE inline void VerifyComma(Node *node, commaNode *u, Context c)
{
     assert(u->exprs != NULL);
     VerifyExprList(u->exprs);
}

PRIVATE inline void VerifyTernary(Node *node, ternaryNode *u, Context c)
{
     VerifyExpr(u->cond);
     VerifyExpr(u->true);
     VerifyExpr(u->false);
}

PRIVATE inline void VerifyArray(Node *node, arrayNode *u, Context c)
{
     VerifyExpr(u->name);
     VerifyExprList(u->dims);
}

PRIVATE inline void VerifyCall(Node *node, callNode *u, Context c)
{
     VerifyExpr(u->name);
     /* u->args is NULL if there are no arguments */
     VerifyExprList(u->args);
}

PRIVATE inline void VerifyInitializer(Node *node, initializerNode *u, Context c)
{
     assert(u->exprs != NULL);
     VerifyExprList(u->exprs);
}

PRIVATE inline void VerifyImplicitCast(Node *node, implicitcastNode *u, Context c)
{
     if (u->expr != NULL)
	  VerifyExpr(u->expr);
     if (u->type != NULL)
	  VerifyType(u->type);
}

/*************************************************************************/
/*                                                                       */
/*                          Statement nodes                              */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void VerifyLabel(Node *node, labelNode *u, Context c)
{
     assert(u->name != NULL);
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifySwitch(Node *node, SwitchNode *u, Context c)
{
     VerifyExpr(u->expr);
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifyCase(Node *node, CaseNode *u, Context c)
{
     VerifyExpr(u->expr);
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifyDefault(Node *node, DefaultNode *u, Context c)
{
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifyIf(Node *node, IfNode *u, Context c)
{
     VerifyExpr(u->expr);
     /* `true' field may be NULL, e.g.,  if (1) {} */
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifyIfElse(Node *node, IfElseNode *u, Context c)
{
     VerifyExpr(u->expr);
     /* `true' field may be NULL, e.g.,  if (1) {} */
     if (u->true)
	  VerifyStmt(u->true);
     /* `false' field may be NULL */
     if (u->false)
	  VerifyStmt(u->false);
}

PRIVATE inline void VerifyWhile(Node *node, WhileNode *u, Context c)
{
     VerifyExpr(u->expr);
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifyDo(Node *node, DoNode *u, Context c)
{
     VerifyExpr(u->expr);
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifyFor(Node *node, ForNode *u, Context c)
{
     /* any field may be NULL */
     if (u->init)
	  VerifyExpr(u->init);
     if (u->cond)
	  VerifyExpr(u->cond);
     if (u->next)
	  VerifyExpr(u->next);
     if (u->stmt)
	  VerifyStmt(u->stmt);
}

PRIVATE inline void VerifyGoto(Node *node, GotoNode *u, Context c)
{
     Node *var;

     var = u->label;
     assert(var != NULL);
     assert(var->typ == Label);
}

PRIVATE inline void VerifyContinue(Node *node, ContinueNode *u, Context c)
{
}

PRIVATE inline void VerifyBreak(Node *node, BreakNode *u, Context c)
{
}

PRIVATE inline void VerifyReturn(Node *node, ReturnNode *u, Context c)
{
     /* expr may be NULL */
     if (u->expr)
	  VerifyExpr(u->expr);
     assert(u->proc);
     assert(u->proc->typ == Proc);
     assert(u->proc->typ == Proc);
}

PRIVATE inline void VerifyBlock(Node *node, BlockNode *u, Context c)
{
     VerifyDeclList(u->decl, Other);
     VerifyStmtList(u->stmts);
}

/*************************************************************************/
/*                                                                       */
/*                             Type nodes                                */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void VerifyPrim(Node *node, primNode *u, Context c)
{
     VerifyTq(u->tq);
     assert(u->basic > 0 && u->basic < MaxBasicType && u->basic != Int_ParseOnly);
}

PRIVATE inline void VerifyTdef(Node *node, tdefNode *u, Context c)
{
     VerifyTq(u->tq);
     assert(u->name != NULL);
     VerifyType(u->type);
}

PRIVATE inline void VerifyPtr(Node *node, ptrNode *u, Context c)
{
     VerifyTq(u->tq);
     VerifyType(u->type);
}

PRIVATE inline void VerifyAdcl(Node *node, adclNode *u, Context c)
{
     assert(u->tq == EMPTY_TQ);
     VerifyType(u->type);
     if (u->dimp->dim)
	  VerifyExpr(u->dimp->dim);	/* can be null */
}

PRIVATE inline void VerifyFdcl(Node *node, fdclNode *u, Context c)
{
     /* allow inlet --- commented by fengmd */
     assert(u->tq == EMPTY_TQ || u->tq == T_INLINE || u->tq == T_PROCEDURE ||
	    u->tq == (T_INLINE | T_PROCEDURE) || u->tq == T_INLET ||
	    u->tq == (T_INLINE | T_INLET));
     VerifyType(u->returns);
     VerifyDeclList(u->args, FormalParm);
}

PRIVATE inline void VerifySdcl(Node *node, sdclNode *u, Context c)
{
     VerifyTq(u->tq);
     assert(u->type != NULL);
     assert(u->type->typ == Sdcl);
     /* u->type->name could be NULL */

     if (SUE_ELABORATED(u->tq)) {
	  VerifyDeclList(u->type->fields, StructField);
     }
}

PRIVATE inline void VerifyUdcl(Node *node, udclNode *u, Context c)
{
     VerifyTq(u->tq);
     assert(u->type != NULL);
     assert(u->type->typ == Udcl);
     /* u->type->name could be NULL */

     if (SUE_ELABORATED(u->tq)) {
	  VerifyDeclList(u->type->fields, StructField);
     }
}

PRIVATE inline void VerifyEdcl(Node *node, edclNode *u, Context c)
{
     VerifyTq(u->tq);
     assert(u->type != NULL);
     /* fix: verify list? */
#if 0
     /* unlike struct/union, there are no empty enum declarations */
     fprintf(stderr, "unexpected node type %d\n", decl->typ);
     assert(FALSE);
#endif
}

/*************************************************************************/
/*                                                                       */
/*                      Other nodes (decls et al.)                       */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void VerifyDecl(Node *node, declNode *u, Context c)
{
     if (c != StructField)
	  assert(u->name != NULL);
     VerifyScAndDq(u->tq);
     VerifyType(u->type);
     if (u->init != NULL) {
	  VerifyExpr(u->init);
	  assert(u->bitsize == NULL);
     }
     if (u->bitsize != NULL) {
	  VerifyExpr(u->bitsize);
	  assert(u->init == NULL);
     }
}

PRIVATE inline void VerifyAttrib(Node *node, attribNode *u, Context c)
{
}

PRIVATE inline void VerifyProc(Node *node, procNode *u, Context c)
{
     assert(u->decl != NULL);
     assert(u->decl->typ == Decl);
     assert(u->decl->u.decl.type != NULL);
     assert(u->decl->u.decl.type->typ == Fdcl);
     VerifyNode(u->decl, Other);

     /* body must be a Block, but could be NULL */
     if (u->body != NULL) {
	  assert(u->body->typ == Block);
	  VerifyStmt(u->body);
     }
}

PRIVATE inline void VerifyText(Node *node, textNode *u, Context c)
{
     /* anything goes */
}

/*************************************************************************/
/*                                                                       */
/*                            Extensions                                 */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void VerifyAsmArg(Node *node, asmargNode *u, Context c)
{
     VerifyExpr(u->constraint);
     VerifyExpr(u->expr);
}

PRIVATE void VerifyAsmArgList(List *list)
{
     ListMarker marker;
     Node *item;

     IterateList(&marker, list);
     while (NextOnList(&marker, (GenericREF) & item)) {
	  assert(item);
	  assert(item->typ == AsmArg);
	  VerifyAsmArg(item, &item->u.AsmArg, Other);
     }
}

PRIVATE inline void VerifyAsm(Node *node, asmNode *u, Context c)
{
     VerifyExpr(u->template);
     if (u->output)
	  VerifyAsmArgList(u->output);
     if (u->input)
	  VerifyAsmArgList(u->input);
     if (u->clobbered)
	  VerifyExprList(u->clobbered);
}

/*************************************************************************/
/*                                                                       */
/*                            Extensions                                 */
/*                                                                       */
/*************************************************************************/

PRIVATE inline void VerifySpawn(Node *node, spawnNode *u, Context c)
{
     if (u->receiver)
	  VerifyExpr(u->receiver);

     VerifyExpr(u->name);
     VerifyExprList(u->args);
}

PRIVATE inline void VerifySync(Node *node, syncNode *u, Context c)
{
}

#if 0
PRIVATE inline void VerifyInlet(Node *node, inletNode * u, Context c)
{
     Node *type;
     assert(u->decl != NULL);
     assert(u->decl->typ == Decl);
     type = u->decl->u.decl.type;
     assert(type != NULL);
     assert(type->typ == Fdcl);
     VerifyNode(u->decl, Other);

     /* body must be a Block, but could be NULL */
     if (u->body != NULL) {
	  assert(u->body->typ == Block);
	  VerifyStmt(u->body);
     }
}
#endif

PRIVATE void VerifyInletCallArg(Node *node)
{
     assert(node);
     assert(IsExpr(node) || (node->typ == Spawn));
     VerifyNode(node, Other);
}

PRIVATE inline void VerifyInletCall(Node *node, inletcallNode *u, Context c)
{
     ListMarker marker;
     Node *item;

     if (u->receiver)
	  VerifyExpr(u->receiver);

     VerifyExpr(u->name);
     assert(u->name->typ == Id);

     IterateList(&marker, u->args);
     while (NextOnList(&marker, (GenericREF) & item)) {
	  VerifyInletCallArg(item);
     }
}


PRIVATE inline void VerifyAbort(Node *node, abortNode *u, Context c)
{
}

PRIVATE inline void VerifySynched(Node *node, synchedNode *u, Context c)
{
}

PRIVATE inline void VerifySpawnCounter(Node *node, spawncounterNode *u, Context c)
{
}

/*************************************************************************/
/*                                                                       */
/*                        Verify et al.                              */
/*                                                                       */
/*************************************************************************/

PRIVATE void VerifyNode(Node *node, Context c)
{
     assert(node);

#define CODE(name, node, union) Verify##name(node, union, c);
     ASTSWITCH(node, CODE)
#undef CODE
}

PRIVATE void VerifyTq(TypeQual tq)
{
     assert(tq == TYPE_QUALS(tq));
}

PRIVATE void VerifyScAndDq(TypeQual tq)
{
     int sc = STORAGE_CLASS(tq);
     int dl = DECL_LOCATION(tq);
     assert(sc == 0 || sc == T_TYPEDEF || sc == T_EXTERN
	    || sc == T_STATIC || sc == T_AUTO || sc == T_REGISTER);
     assert(dl == T_TOP_DECL || dl == T_BLOCK_DECL || dl == T_FORMAL_DECL ||
	    dl == T_SU_DECL || dl == T_ENUM_DECL);
     assert(TYPE_QUALS(tq) == 0);
}

PRIVATE void VerifyType(Node *type)
{
     assert(type);
     assert(IsType(type));
     VerifyNode(type, Other);
}

PRIVATE void VerifyDeclList(List *list, Context c)
{
     ListMarker marker;
     Node *decl;

     IterateList(&marker, list);
     while (NextOnList(&marker, (GenericREF) & decl)) {
	  assert(decl);

	  switch (c) {
	      case TopLevel:
		   assert(IsDecl(decl));
		   break;

	      case FormalParm:
		   assert(decl->typ == Decl ||
			  IsType(decl));
		   break;

	      default:
		   /* allow local proc decls!!! KHR */
		   /* assert(IsDecl(decl) && decl->typ != Proc); */
		   break;
	  }

	  VerifyNode(decl, c);
     }
}

PRIVATE void VerifyExpr(Node *node)
{
     assert(node);
     assert(IsExpr(node));
     VerifyNode(node, Other);
}

PRIVATE void VerifyExprList(List *list)
{
     ListMarker marker;
     Node *item;

     IterateList(&marker, list);
     while (NextOnList(&marker, (GenericREF) & item))
	  VerifyExpr(item);
}

PRIVATE void VerifyTypeOrExpr(Node *node)
{
     assert(node);
     assert(IsType(node) || IsExpr(node));
     VerifyNode(node, Other);
}

PRIVATE void VerifyStmt(Node *node)
{
     assert(node);
     assert(IsStmt(node));
     VerifyNode(node, Other);
}

PRIVATE void VerifyStmtList(List *list)
{
     ListMarker marker;
     Node *item;

     IterateList(&marker, list);
     while (NextOnList(&marker, (GenericREF) & item))
	  if (item)		/* NULL statements are legal */
	       VerifyStmt(item);
}
